@@ -165,6 +165,7 @@ WIRE_VERSION_CHECK_DECLS (4)
 WIRE_VERSION_CHECK_DECLS (5)
 WIRE_VERSION_CHECK_DECLS (6)
 WIRE_VERSION_CHECK_DECLS (7)
+WIRE_VERSION_CHECK_DECLS (8)
 
 #undef WIRE_VERSION_CHECK_DECLS
 
@@ -197,6 +198,9 @@ test_framework_skip_if_no_compressors (void);
 int
 test_framework_skip_if_compressors (void);
 
+int
+test_framework_skip_if_no_failpoint (void);
+
 void
 test_framework_resolve_path (const char *path, char *resolved);
 
